// Title:   MovingLeastSquares.cpp
// Created: Thu Sep 25 15:11:17 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "MovingLeastSquares.h"

float MovingLeastSquares::evaluateFunction(float t) {

	int i;
	float value = 0.0f;
	float dotProd = 0.0f;
	Vector3D pi, vec;

	nbHood->setSourcePoint(sourcePoint);
	nbHood->setNofNeighbours(numberOfNeighbours);

	for (i = 0; i < numberOfNeighbours; i++) {
		pi = nbHood->getNeighbourPosition(i);
		vec = pi - sourcePoint - t*this->normal;
		dotProd = Vector3D::dotProduct(this->normal, vec);
		value += dotProd * dotProd * exp(vec.getSquaredLength() / (_h*_h));
	}

	return value;
}


float MovingLeastSquares::evaluateDerivative(float t) {
	int i;
	float value = 0.0f;
	float dotProd = 0.0f;
	Vector3D pi, vec;

	nbHood->setSourcePoint(sourcePoint);
	nbHood->setNofNeighbours(numberOfNeighbours);

	for (i = 0; i < numberOfNeighbours; i++) {
		pi = nbHood->getNeighbourPosition(i);
		vec = pi - sourcePoint - t*this->normal;
		dotProd = Vector3D::dotProduct(this->normal, vec);
		value += dotProd * (1 + (dotProd*dotProd)/ (_h*_h)) * exp(vec.getSquaredLength() / (_h*_h));
	}

	value *= 2;

	return value;
}


float MovingLeastSquares::evaluateFunction(Vector3D q) {

	int i;
	float value = 0.0f;
	float dotProd = 0.0f;
	Vector3D pi, vec;

	nbHood->setSourcePoint(sourcePoint);
	nbHood->setNofNeighbours(numberOfNeighbours);

	for (i = 0; i < numberOfNeighbours; i++) {
		pi = nbHood->getNeighbourPosition(i);
		vec = pi - q;
		dotProd = Vector3D::dotProduct(this->normal, vec);
		value += dotProd * dotProd * exp(vec.getSquaredLength() / (_h*_h));
	}

	return value;
}


Vector3D MovingLeastSquares::evaluateDerivative(Vector3D q) {

	// Imporant note: conjugate gradient not implemented yet!
	// (all functions needed are in this class)

	return NULL;
}


float MovingLeastSquares::evaluateFunction1dim(float x) {
	float f;
	Vector3D xt;

	xt = _p + x * _xi;
	f = (*this.*_f)(xt);

	return f;
}


inline float MovingLeastSquares::evaluateDerivative1dim(float x) {

	float df1 = 0.0;
	Vector3D xt, df;

	xt = _p + x * _xi;
	df = (*this.*_df)(xt);
	df1 = Vector3D::dotProduct(df, _xi);

	return df1;
}


float MovingLeastSquares::minimizeBrent(float ax, float bx, float cx, float (MovingLeastSquares::*f)(float), float (MovingLeastSquares::*df)(float), float tol, float *xmin) {

	// parts of the code from Numerical Recipes in C, 2nd Edition, pp 405ff

	int iter, ok1, ok2; 	// Will be used as flags for whether proposed steps are acceptable or not.
	float a, b, d, d1, d2, du, dv, dw, dx, e = 0.0;
	float fu, fv, fw, fx, olde, tol1, tol2, u, u1, u2, v, w, x, xm;

	a = (ax < cx ? ax : cx);
	b = (ax > cx ? ax : cx);
	x = w = v = bx;
	fw = fv = fx = (*this.*f)(x);
	dw = dv = dx = (*this.*df)(x);

	// All our housekeeping chores are doubled by the necessity of moving
	// derivative values around as well as function values.

	for (iter = 1; iter <= ITMAX; iter++) {
		xm = 0.5 * (a + b);
		tol1 = tol * fabs(x) + ZEPS;
		tol2 = 2.0 * tol1;

		if (fabs(x - xm) <= (tol2 - 0.5*(b - a))) {
			*xmin=x;
			return fx;
		}

		if (fabs(e) > tol1) {	// Initialize these d's to an out-of-bracket value.
			d1 = 2.0 * (b-a);
			d2 = d1;
			if (dw != dx) {
				d1 = (w - x)*dx/(dx - dw); // Secant method with one point.
			}
			if (dv != dx) {
				d2 = (v - x)*dx/(dx - dv); // And the other.
			}

			// Which of these two estimates of d shall we take? We will insist that they be within
			// the bracket, and on the side pointed to by the derivative at x:

			u1 = x + d1;
			u2 = x + d2;
			ok1 = (a - u1)*(u1 - b) > 0.0 && dx*d1 <= 0.0;
			ok2 = (a - u2)*(u2 - b) > 0.0 && dx*d2 <= 0.0;
			olde = e;	// Movement on the step before last.
			e = d;

			// Take only an acceptable d, and if both are acceptable, then take the smallest one.

			if (ok1 || ok2) {
				if (ok1 && ok2) {
					d = (fabs(d1) < fabs(d2) ? d1 : d2);
				}
				else if (ok1) {
					d = d1;
				}
				else {
					d = d2;
				}
				if (fabs(d) <= fabs(0.5 * olde)) {
					u = x + d;
					if (u-a < tol2 || b-u < tol2) {
						d = SIGN(tol1, xm - x);
					}
				}
				else {	//  Bisect, not golden section.
					d = 0.5 * (e = (dx >= 0.0 ? a - x : b - x));
					//Decide which segment by the sign of the derivative.
				}
			}
			else {
				d = 0.5 * (e = (dx >= 0.0 ? a - x : b - x));
			}
		}
		else {
			d = 0.5 * (e = (dx >= 0.0 ? a - x : b - x));
		}
		if (fabs(d) >= tol1) {
			u = x + d;
			fu = (*this.*f)(u);
		}
		else {
			u = x + SIGN(tol1,d);
			fu = (*this.*f)(u);

			// If the minimum step in the downhill direction takes us uphill, then we are done.
			if (fu > fx) {
				*xmin = x;
				return fx;
			}
		}
		du = (*this.*df)(u);	// Now all the housekeeping, sigh.

		if (fu <= fx) {
			if (u >= x) {
				a = x;
			}
			else {
				b = x;
			}
			MOV3(v,fv,dv,w,fw,dw)
			MOV3(w,fw,dw,x,fx,dx)
			MOV3(x,fx,dx,u,fu,du)
		}
		else {
			if (u < x) {
				a = u;
			}
			else {
				b = u;
			}
			if (fu <= fw || w == x) {
				MOV3(v,fv,dv,w,fw,dw)
				MOV3(w,fw,dw,u,fu,du)
			}
			else if (fu < fv || v == x || v == w) {
				MOV3(v,fv,dv,u,fu,du)
			}
		}
	}
	qDebug("Too many iterations in routine minimizeBrent");
	return 0.0; 	// Never get here.
}


void MovingLeastSquares::bracketingMinimun(float *ax, float *bx, float *cx, float *fa, float *fb,float *fc, float (MovingLeastSquares::*f)(float)) {

	float ulim, u, r, q, fu, dum;
	*fa = (*this.*f)(*ax);
	*fb = (*this.*f)(*bx);

	if (*fb > *fa) { 	// Switch roles of a and b so that we can go downhill in the direction from a to b.
		SHFT(dum,*ax,*bx,dum)
		SHFT(dum,*fb,*fa,dum)
	}

	*cx = (*bx) + GOLD * (*bx - *ax); 	// First guess for c.
	*fc = (*this.*f)(*cx);

	while (*fb > *fc) {	// Keep returning here until we bracket.
		r = (*bx - *ax) * (*fb - *fc);	// Compute u by parabolic extrapolation from a, b, c. TINY is used to prevent any possible division by zero.
		q = (*bx - *cx) * (*fb - *fa);
		u = (*bx) - ((*bx - *cx) * q - (*bx - *ax) * r) / (2.0 * SIGN(FMAX(fabs(q - r),TINY),q - r));
		ulim = (*bx) + GLIMIT * (*cx - *bx);
		// We wont go farther than this. Test various possibilities:
		if ((*bx - u) * (u - *cx) > 0.0) { // Parabolic u is between b and c: try it.
			fu = (*this.*f)(u);

			if (fu < *fc) {		// Got a minimum between b and c.
				*ax = (*bx);
				*bx = u;
				*fa = (*fb);
				*fb = fu;
				return;
			}
			else if (fu > *fb) { // Got a minimum between between a and u.
				*cx = u;
				*fc = fu;
				return;
			}
			u = (*cx) + GOLD * (*cx - *bx); // Parabolic fit was no use. Use default magnification.
			fu = (*this.*f)(u);
		}
		else if ((*cx - u) * (u - ulim) > 0.0) { 	// Parabolic fit is between c and its allowed limit.
			fu = (*this.*f)(u);
			if (fu < *fc) {
				SHFT(*bx,*cx,u,*cx+GOLD*(*cx - *bx))
				SHFT(*fb,*fc,fu,(*this.*f)(u))
			}
		}
		else if ((u - ulim)*(ulim - *cx) >= 0.0) { // Limit parabolic u to maximum allowed value.
			u = ulim;
			fu = (*this.*f)(u);
		}
		else { 	// Reject parabolic u, use default magnification.
			u = (*cx) + GOLD * (*cx - *bx);
			fu = (*this.*f)(u);
		}
		SHFT(*ax,*bx,*cx,u);	// Eliminate oldest point and continue.
		SHFT(*fa,*fb,*fc,fu);
	}
}


void MovingLeastSquares::dlinmin(Vector3D p, Vector3D xi, float *fret, float (MovingLeastSquares::*f)(Vector3D), Vector3D (MovingLeastSquares::*df)(Vector3D)) {

	float xx, xmin, fx, fb, fa, bx, ax;

	// Define the global variables.
	_f = f;
	_df = df;
	_p = p;
	_xi = xi;

	ax = 0.0;	// Initial guess for brackets.
	xx = 1.0;
	bracketingMinimun(&ax, &xx, &bx, &fa, &fx, &fb, &MovingLeastSquares::evaluateFunction1dim);
	*fret = minimizeBrent(ax, xx, bx, &MovingLeastSquares::evaluateFunction1dim, &MovingLeastSquares::evaluateDerivative1dim, TOL, &xmin);

	xi *= xmin; // Construct the vector results to return.
	p *= xi;

}


void MovingLeastSquares::conjugateGradient(Vector3D p, float ftol, int *iter, float *fret, float (MovingLeastSquares::*f)(Vector3D), Vector3D (MovingLeastSquares::*df)(Vector3D)) {

	int j, its;
	float gg, gam, fp, dgg;
	Vector3D g, h, xi;
	fp = (*this.*f)(p);	// Initializations.
	xi = (*this.*df)(p);

	g = -xi;
	xi = h = g;

	for (its = 1; its <= ITMAX2; its++) { // Loop over iterations.
		*iter = its;
		dlinmin(p, xi, fret, f, df);	// Next statement is the normal return:

		if (2.0 * fabs(*fret - fp) <= ftol * (fabs(*fret) + fabs(fp) + EPS)) {
			return;
		}
		fp = *fret;
		xi = (*this.*df)(p);
		dgg = gg = 0.0;
		for (j = 0; j < 3; j++) {
			gg += g[j] * g[j];
		//	dgg += xi[j] * xi[j];	// This statement for Fletcher-Reeves.
			dgg += (xi[j] + g[j]) * xi[j]; // This statement for Polak-Ribiere.
		}
		if (gg == 0.0) { // Unlikely. If gradient is exactly zero then we are already done.
			return;
		}
		gam = dgg / gg;
		g = -xi;
		xi = h = g + gam * h;
	}

	qDebug("Too many iterations in conjugateGradient");
}


void MovingLeastSquares::solve(Vector3D sourcePoint) {

	float xmin;
	Vector3D p;
	this->sourcePoint = sourcePoint;
	_t = minimizeBrent(-_h/2, 0.0f, _h/2, &MovingLeastSquares::evaluateFunction, &MovingLeastSquares::evaluateDerivative, TOL, &xmin);

	// Imporant note: conjugate gradient not implemented yet!
	// Vector3D evaluateDerivative(Vector3D q) has to be calculated
	// float fret;
	// int iter;
	// p = _sourcePoint + _t*this->normal;
	// conjugateGradient(p, TOL, &iter, &fret, evaluateFunction, evaluateDerivative);
}


void MovingLeastSquares::projection() { 

	int     i, j;

	gsl_matrix *X, *cov;	// weighted least squares fit
	gsl_vector *y,*c,*w;
	double chisquare;
	
	Vector3D u_direction, v_direction;
	float z_coord;

	int polynomialVars;
	// for polynomial of fourth or third degree
	if (this->polynomial == 4) {
		polynomialVars = 15;
		if (this->numberOfNeighbours < 15) {
			// set number of neighbours also to minimum 15
			this->numberOfNeighbours = 15;
		}
	}
	else {
		this->polynomial = 3;
		polynomialVars = 10;
		if (this->numberOfNeighbours < 10) {
			// set number of neighbours also to minimum 10
			this->numberOfNeighbours = 10;
		}
	}

	this->nbHood->setNofNeighbours(this->numberOfNeighbours);

	X = gsl_matrix_alloc(this->numberOfNeighbours, polynomialVars);
	y = gsl_vector_alloc(this->numberOfNeighbours);
	w = gsl_vector_alloc(this->numberOfNeighbours);

	c = gsl_vector_alloc(polynomialVars);
	cov = gsl_matrix_alloc(polynomialVars, polynomialVars);

	for (j = 0; j < this->numberOfSurfels; j++) {

		PlaneFit plane;

		if (this->recalculateNormals) {
			// Weighted Least Squares
			plane.doPlaneFitAdaptiveWeight(this->nbHood, this->positions[j], this->numberOfNeighbours);
			this->normal = plane.getNormal();
		}
		else {
			// take normal from surfelcollection
			this->normal = this->normals[j];
		}
		
		if (approximation) {
			if (this->recalculateNormals) {
				if (this->normal[2] != 0.0f) {
					z_coord = (float) -1. / normal[2] * ( (normal[0] * 2 + normal[1] * 2 + plane.getD()) - (normal[0] + normal[1] + plane.getD()));
				}
			}
			else {
				if (this->normal[2] != 0.0f) {
					z_coord = -(this->normal[0] + this->normal[1]) / this->normal[2];
				}
			}
		}
		else {
			this->sourcePoint = this->positions[j];
			this->nbHood->setNofNeighbours(this->numberOfNeighbours);
			this->_t = 0.0f;
			this->_h = this->nbHood->getSquaredDistance(this->numberOfNeighbours - 1) / 3.0;
			this->solve(this->positions[j]);
			
			if (this->normal[2] != 0.0f) {
				z_coord = -1. / this->normal[2] * ( (this->normal[0] * 2 + this->normal[1] * 2 + this->_t) - (this->normal[0] + this->normal[1] + this->_t));
			}
		}
		
		if (Vector3D::dotProduct(this->normal, this->normals[j]) < 0.0f) {
			this->normal *= -1;
		}

		if (this->normal[2] == 0.0f) {
			u_direction = Vector3D(0,0,1);
		}
		else {
			u_direction = Vector3D(1.,1.,z_coord);
		}
		v_direction = Vector3D(0.,0.,0.);
		u_direction.normalize();
		v_direction = Vector3D::crossProduct(u_direction, this->normal);
		v_direction.normalize();
		this->nbHood->setSourcePoint(this->positions[j]);

		for (i = 0; i < this->numberOfNeighbours; i++) {

			Vector3D point = this->nbHood->getNeighbourPosition(i);
			float fi = Vector3D::dotProduct(point - this->positions[j], this->normal);
			float ui = Vector3D::dotProduct(point - this->positions[j], u_direction);
			float vi = Vector3D::dotProduct(point - this->positions[j], v_direction);

			// bivariate polynom of degree four:
			// g(x,y) = x^4 + x^3*y + x^2*y^2 + x*y^3 + y^4 + 
			//			x^3 + y^3 + x^2 + x^2*y + y^2 + y^2*x + x + y + x*y + D;
			if (this->polynomial == 4) {
				gsl_matrix_set (X, i,  0, ui*ui*ui*ui);
				gsl_matrix_set (X, i,  1, ui*ui*ui*vi);
				gsl_matrix_set (X, i,  2, ui*ui*vi*vi);
				gsl_matrix_set (X, i,  3, ui*vi*vi*vi);
				gsl_matrix_set (X, i,  4, vi*vi*vi*vi);
				gsl_matrix_set (X, i,  5, ui*ui*ui);
				gsl_matrix_set (X, i,  6, vi*vi*vi);
				gsl_matrix_set (X, i,  7, ui*ui);
				gsl_matrix_set (X, i,  8, ui*ui*vi);
				gsl_matrix_set (X, i,  9, vi*vi);
				gsl_matrix_set (X, i, 10, vi*vi*ui);
				gsl_matrix_set (X, i, 11, vi*ui);
				gsl_matrix_set (X, i, 12, ui);
				gsl_matrix_set (X, i, 13, vi);
				gsl_matrix_set (X, i, 14, 1);
				gsl_vector_set (y, i, fi);
			}
			// bivariate polynom of degree three:
			// g(x,y) = x^3 + y^3 + x^2 + x^2*y + y^2 + y^2*x + x + y + x*y + D;
			else if (this->polynomial == 3) {
				gsl_matrix_set (X, i,  0, ui*ui*ui);
				gsl_matrix_set (X, i,  1, vi*vi*vi);
				gsl_matrix_set (X, i,  2, ui*ui);
				gsl_matrix_set (X, i,  3, ui*ui*vi);
				gsl_matrix_set (X, i,  4, vi*vi);
				gsl_matrix_set (X, i,  5, vi*vi*ui);
				gsl_matrix_set (X, i,  6, vi*ui);
				gsl_matrix_set (X, i,  7, ui);
				gsl_matrix_set (X, i,  8, vi);
				gsl_matrix_set (X, i,  9, 1);
				gsl_vector_set (y, i, fi);
			}
			
			// calculate parameter h dynamically (-ln 0.05 =~ 3) 
			double factor = this->nbHood->getSquaredDistance(this->numberOfNeighbours - 1) / 3.0;

			float distance = Vector3D::distance(point, this->positions[j]);
			float weight =  exp(-(distance*distance)/(factor*factor));
			gsl_vector_set (w, i,weight);
		}

		gsl_matrix_const_view X_view = gsl_matrix_const_submatrix(X, 0,0, this->numberOfNeighbours, polynomialVars);
		gsl_vector_const_view y_view = gsl_vector_const_subvector(y, 0, this->numberOfNeighbours);
		gsl_vector_const_view w_view = gsl_vector_const_subvector(w, 0, this->numberOfNeighbours);

		// solving least squares equation
		gsl_multifit_linear_workspace *workspace = gsl_multifit_linear_alloc (this->numberOfNeighbours, polynomialVars);
		gsl_multifit_wlinear(&X_view.matrix, &w_view.vector, &y_view.vector, c, cov, &chisquare, workspace);
		
		this->newPositions[j] = this->positions[j] + this->normal*gsl_vector_get(c,polynomialVars - 1);

		// getting the normal.
		if (this->adjustNormals) {

			float dds = gsl_vector_get(c,polynomialVars - 3);
			float ddt = gsl_vector_get(c,polynomialVars - 2);

			Vector3D dxds = u_direction + Vector3D::crossProduct(u_direction,v_direction) * dds;
			Vector3D dxdt = v_direction + Vector3D::crossProduct(u_direction,v_direction) * ddt;

			this->newNormals[j] = Vector3D::crossProduct(dxdt, dxds);
			 
			this->newNormals[j].normalize();
		}
		
		gsl_multifit_linear_free(workspace);
	}

	gsl_matrix_free(X);
	gsl_matrix_free(cov);
	gsl_vector_free(y);
	gsl_vector_free(w);
	gsl_vector_free(c);
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
