// Title:   PlaneFit.cpp
// Created: Thu Sep 25 15:10:20 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "PlaneFit.h"


PlaneFit::PlaneFit() {
	_normal = Vector3D(0.,0.,0.);
	_isvalid = false;
}


void PlaneFit::doPlaneFit(NeighbourHood *nbHood, const Vector3D& source, int nofNeighbours, float param) {

	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;

	int i;
	float *weights;
	float totalweight;
	
	// matrix init:
	data_matrix = gsl_matrix_calloc(3,nofNeighbours);
	transposed_data_matrix = gsl_matrix_calloc(nofNeighbours, 3);
	covariance_matrix = gsl_matrix_calloc(3,3);
	eigen_vectors = gsl_matrix_calloc(3,3);
	eigen_values = gsl_vector_calloc(3);
	workspace = gsl_eigen_symmv_alloc(3);

	// other
	if (!nbHood) {
		return;
	}

	nbHood->setNofNeighbours(nofNeighbours);
	nbHood->setSourcePoint(source);
	
	weights = new float[nofNeighbours];
	totalweight = 0;
	_center = Vector3D(0.,0.,0.);
	
	if (param == 0) {
		return;
	}
		
	for (i = 0; i < nofNeighbours; i++) {
	
		Vector3D vec = nbHood->getNeighbourPosition(i);
		
		// get weights
		float current_distance = Vector3D::distance(vec, source);
		weights[i] = exp(-(current_distance*current_distance) / param);
		totalweight += weights[i];
		// get data matrix entries
		gsl_matrix_set(data_matrix,0,i,vec[0]);
		gsl_matrix_set(data_matrix,1,i,vec[1]);
		gsl_matrix_set(data_matrix,2,i,vec[2]);
		// update center
		_center += vec * weights[i];
	}

	if (totalweight == 0) {
		return;
	}

	_center /= totalweight;
	
	// set up data matrix
	for (i = 0; i < nofNeighbours; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - _center[0])*weights[i]);
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - _center[1])*weights[i]);
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - _center[2])*weights[i]);
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);
	// create matrix views
	gsl_matrix_const_view data_matrix_view = gsl_matrix_const_submatrix(data_matrix, 0,0, 3, nofNeighbours);
	gsl_matrix_const_view transposed_data_matrix_view = gsl_matrix_const_submatrix(transposed_data_matrix, 0,0, nofNeighbours, 3);
	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, &data_matrix_view.matrix, &transposed_data_matrix_view.matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);
	
	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	_normal[0] = gsl_matrix_get(eigen_vectors,0,0);
	_normal[1] = gsl_matrix_get(eigen_vectors,1,0);
	_normal[2] = gsl_matrix_get(eigen_vectors,2,0);
	_normal.normalize();

	_d = -1.*(_normal[0]*_center[0]+_normal[1]*_center[1]+_normal[2]*_center[2]);
	_isvalid = true;

	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);
}


void PlaneFit::doPlaneFitAdaptiveWeight(NeighbourHood *nbHood, const Vector3D& source, int nofNeighbours) {

	float param;
	
	if (!nbHood) {
		return;
	}

	nbHood->setNofNeighbours(nofNeighbours);
	nbHood->setSourcePoint(source);
	
	// calc param adaptive ( - ln 0.05 =~ 3) 
	param = nbHood->getSquaredDistance(nofNeighbours - 1) / 3.0;

	doPlaneFit(nbHood,source,nofNeighbours,param);
}


void PlaneFit::getEigenValues(float eigenvals[3]) const {
	eigenvals[0] = _eigenvalues[0];
	eigenvals[1] = _eigenvalues[1];
	eigenvals[2] = _eigenvalues[2];
}


Vector3D PlaneFit::getNormal() const {
	return _normal;
}


Vector3D PlaneFit::getCenter() const {
	return _center;
}


float PlaneFit::getD() const  {
	return _d;
}


void PlaneFit::reset() {
	_normal = Vector3D(0.,0.,0.);
}


float PlaneFit::getDistance(Vector3D source) const {
	return Vector3D::dotProduct((source - _center),_normal);
}


bool PlaneFit::isValid() const {
	return _isvalid;
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
