// Title:   Pool.h
// Created: Thu Sep 25 15:10:17 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __POOL_H_
#define __POOL_H_

#include <vector>


template<class T> class chunk{
public:
    chunk(int n) { items = new T[n]; }
    ~chunk() { delete[] items; }

public:
    chunk<T> *  next;
    T *  items;
};


/**
  * This template class is used to allocate items in chunks. This decreases the allocation
  * overhead. Chunks are linked with a pointer to the next chunk.
  */
template<class T> class pool {

public:

  /**
	* @param cs
	*			defines the chunk size.
	* @param nc
	*			defines the number of chunks.
	* @param isGrowable
	*			determines whether pool is growable or not.
	*/
	pool(int cs, int nc, bool grow = true) : freeList(cs*nc) {
		chunkSize = cs;
        isGrowable = grow;
        poolSize = 0;
        allocatedCount = 0;

        chunkp = NULL;
        for(int i=0; i<nc; i++)
            allocChunk();
    }
   
	~pool() {
        while(chunkp != NULL) {
            chunk<T> *  oldp = chunkp;
            chunkp = chunkp->next;
            delete oldp;
        }
    }

   /**
	* Allocate item.
	*/	
    T *  alloc() {
        if (freeList.size() == 0) {
            if (!isGrowable)
                return NULL;

            allocChunk();
        }

        allocatedCount++;
		T *element=freeList.back();
        freeList.pop_back();
		return element;
    }
   
   /**
	* Deallocate item.
	*/	
    void dealloc(T *t) {
        allocatedCount--;
        freeList.push_back(t);
    }

   /**
	* Reset pool.
	*/	
    void reset() {
        freeList.reset();
        chunk<T> *  cp = chunkp;
        while(cp != NULL) {
            for(int i=0; i<chunkSize; i++)
                freeList.add(&cp->items[i]);
        }
        allocated_count = 0;
    }

   /**
	* @return 
	*		number of allocated items.
	*/	
    int getAllocatedCount() {
        return allocatedCount;
    }


protected:
    pool() { };

    void allocChunk() {
        chunk<T> *  new_chunkp = new chunk<T>(chunkSize);
        poolSize += chunkSize;
        for(int i=0; i<chunkSize; i++)
            freeList.push_back(&(new_chunkp->items[i]));
        new_chunkp->next = chunkp;
        chunkp = new_chunkp;
    }


private:
    chunk<T> *  chunkp;
    int chunkSize;
    std::vector<T *>  freeList;
	bool isGrowable;
    int poolSize;
    int allocatedCount;
};

#endif  // __POOL_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
