// Title:   VolGridParticleSim.cpp
// Created: Thu Sep 25 15:10:03 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "VolGridParticleSim.h"


void VolGridParticleSimulation::build() {

	Object           *object;

	this->scene = Scene::getInstance();
	object = new Object("Particle Simulation");
	this->surfels = object->getSurfelCollection();

	SurfelInterface::PropertyDescriptor pd = SurfelInterface::BaseSurfel_PropertyDescriptor;
	surfels->setPropertyDescriptor(pd);
	surfels->reserve(this->numSamples);

	this->rest = 0.0f;
	this->cellCounter = 0;
	
	// count active cells
	VolGridActiveCellCounter *counter = new VolGridActiveCellCounter(this->grid);
	counter->count();
	qDebug("Number of active cells: %i", counter->nofActiveCells);
	this->numSamplesPerCell = ((float) this->numSamples / (float) counter->nofActiveCells);

	// calculate variation
	VolGridVariation *var = new VolGridVariation(this->grid, counter->nofActiveCells);
	var->calculateVariation();
	this->variation = var->variation;
	this->variationAvg = var->variationAvg;
	
	// particle simulation.
	this->grid->traverse(this);

	Object *active;
	active = scene->getActiveObject();
	if ((active != NULL) && (active->getName() == "Particle Simulation")) {
		scene->removeObject(active, false);
	}

	scene->addObject(object);
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	delete counter;
	delete var;
}


bool VolGridParticleSimulation::iterate(OcTreeTraversalNode *ocTravNode) {

	int i, j;
	SurfelInterface *surfel;
	Vector3D gradient;
	float f[8];
	int type = 0;
	bool noActiveCell = false;
	
	VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

	// Determine if child is a leaf node at maximum depth.
	if (travNode->depth == this->grid->getMaxDepth() && ((VolGridNode *) (travNode->node))->nodeType == LEAF &&
		((VolGridNode *) (travNode->node))->valid == true) {
	
		for (j = 0; j < 8; j++) {
			f[j] = *(node->distanceValues[j]);
		}

		if (f[0] > 0.0) type |=   1;
		if (f[1] > 0.0) type |=   2;
		if (f[2] > 0.0) type |=   4;
		if (f[3] > 0.0) type |=   8;
		if (f[4] > 0.0) type |=  16;
		if (f[5] > 0.0) type |=  32;
		if (f[6] > 0.0) type |=  64;
		if (f[7] > 0.0) type |= 128;

		if (type == 0 || type == 255) {
			return true;
		}

		Vector3D pos;
		
		for (i = 0; i < (int) ((this->numSamplesPerCell*((1 - this->variationInfluence) +
			                    this->variationInfluence*(this->variation[cellCounter]/variationAvg))) + this->rest); i++) {
			
			// note that the surfel is not placed "on" the surface for faster calulcation!
			int s;
			for (s = 0; s < 3; s++) {
				pos[s] = 2 * travNode->length * ((float) rand() / RAND_MAX);
			}

			float h[6];

			// side length of a cell
			float a = 2 * travNode->length;
			float p, q, r;
			float g0, g1, g2, g3;
			q = pos[1];
			r = pos[2];

			float h0 = (1-q/a)*f[0] + (q/a) * f[2];
			float h1 = (1-q/a)*f[4] + (q/a) * f[6];
			h[0] = (1-(r/a))*h0 + (r/a)*h1;

			p = 2 * travNode->length;
			q = pos[1];
			r = pos[2];
			h0 = (1-q/a)*f[1] + (q/a) * f[3];
			h1 = (1-q/a)*f[5] + (q/a) * f[7];
			h[1] = (1-r/a)*h0 + (r/a)*h1;
			
			p = pos[0];
			r = pos[2];
			g0 = (1-p/a)*f[0] + (p/a) * f[1];
			g2 = (1-p/a)*f[4] + (p/a) * f[5];
			h[2] = (1-(r/a))*g0 + (r/a)*g2;

			p = pos[0];
			r = pos[2];
			q = 2 * travNode->length;
			g1 = (1-p/a)*f[2] + (p/a) * f[3];
			g3 = (1-p/a)*f[6] + (p/a) * f[7];
			h[3] = (1-r/a)*g1 + (r/a)*g3;

			p = pos[0];
			q = pos[1];
			g0 = (1-p/a)*f[0] + (p/a) * f[1];
			g1 = (1-p/a)*f[2] + (p/a) * f[3];
			h[4] = (1-q/a)*g0 + (q/a) * g1;

			p = pos[0];
			q = pos[1];
			r = 2 * travNode->length;
			g2 = (1-p/a)*f[4] + (p/a) * f[5];
			g3 = (1-p/a)*f[6] + (p/a) * f[7];
			h[5] = (1-q/a)*g2 + (q/a) * g3;

			gradient[0] = h[1] - h[0];
			gradient[1] = h[3] - h[2];
			gradient[2] = h[5] - h[4];
		
			
			for (s = 0; s < 3; s++) {
				pos[s] += (travNode->center[s] - travNode->length);
			}
			
			surfel = this->surfels->addSurfel(false);
			scene->setSurfelPosition(surfel, pos, false);
			gradient.normalize();
			scene->setSurfelNormal(surfel, gradient, false);
			scene->setSurfelRadius(surfel, (2*travNode->length), false);
			scene->setSurfelDiffuseColor(surfel, qRgb(240, 240, 240), false);
			scene->setSurfelSpecularColor(surfel, qRgb(240, 240, 240), false);
			scene->setSurfelFlags(surfel, SurfelInterface::NO_FLAGS, false);
		}
		
		this->rest = (this->numSamplesPerCell*((1 - this->variationInfluence) + this->variationInfluence*(this->variation[cellCounter]/variationAvg))) + this->rest -
		             ((int) ((this->numSamplesPerCell*((1 - this->variationInfluence) + this->variationInfluence*(this->variation[cellCounter]/variationAvg))) + this->rest));
		cellCounter++; 
	}

	return true;
}


void VolGridParticleSimulation::repel() {

	Scene            *scene;
	Object           *object;
	NeighbourHood    *modelSearch;      // search structure
	Vector3D         *modelPositions;   // positions of surfels
	SurfelCollection *surfels;
	SurfelInterface  *surfel;
	unsigned int     i, j, k;

	Vector3D		 *forces;
	Vector3D		 direction;
	Vector3D		 gradient;
	Vector3D		 newPos;
	float            forceMagnitude;

	QString          string;
	QString          filePath;
	QString          fileFormat;
	float			 *forceScales;
	float            variationScale;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();
		if (nofSurfels == 0) {
			return;
		}
					
		VolGridTraversalNode *travNode = (VolGridTraversalNode *)this->grid->nodeFactory->newTraversalNode();
		modelPositions = new Vector3D [nofSurfels];
		forces = new Vector3D [nofSurfels];
		forceScales = new float [nofSurfels];
		
		// create search structure for nearest point queries
		for (i = 0, surfel = surfels->getFirstSurfel(); i < surfels->getNofSurfels(); i++, surfel = surfels->getNextSurfel()) {
			modelPositions[i] = surfel->getPosition();
		}

		// kd-Tree is generally faster for few iterations.
		// query grid yields better results - but the size of the queryradius is critical
		modelSearch = new NeighbourHood (modelPositions, surfels->getNofSurfels());
		modelSearch->setNofNeighbours (this->numNeighbours); 
		
		for (k = 0; k < this->numIterations; k++) {
			
			modelSearch->setPositions (modelPositions, surfels->getNofSurfels());
			modelSearch->setNofNeighbours (this->numNeighbours); 

			for (i = 0, surfel = surfels->getFirstSurfel(); i < surfels->getNofSurfels(); i++, surfel = surfels->getNextSurfel()) {
				Vector3D boundingBox = Vector3D (2 * this->grid->length, 2 * this->grid->length, 2 * this->grid->length);
				Vector3D bBoxMinSquare = this->grid->center - Vector3D(this->grid->length, this->grid->length, this->grid->length);
				Vector3D normalizedPos = (surfel->getPosition() - bBoxMinSquare) / (boundingBox);
				this->grid->getCellForPoint(normalizedPos, travNode);
				variationScale = this->grid->computeSurfaceVariation(travNode->center);
				forceScales[i] = (this->variationAvg / variationScale);
			}

			
			for (i = 0, surfel = surfels->getFirstSurfel(); i < surfels->getNofSurfels(); i++, surfel = surfels->getNextSurfel()) {
				
				modelSearch->setSourcePoint (surfel->getPosition());
				
				forces[i].makeZero();
				
				for (j = 1; j < this->numNeighbours; j++) {
					
					direction = surfel->getPosition() - modelSearch->getNeighbourPosition(j);
					direction.normalize();
					forceMagnitude = sqrt (modelSearch->getSquaredDistance (this->numNeighbours - 1)) - 
									 sqrt (modelSearch->getSquaredDistance (j));
					forceMagnitude *= forceScales[modelSearch->getNeighbourPositionIndex(j)];
					forceMagnitude *= this->forceScale;
					
					if (forceMagnitude < 0) {
						forceMagnitude = 0;
					}
									
					forces[i] += forceMagnitude * direction;
				}
			}		
			

			for (i = 0, surfel = surfels->getFirstSurfel(); i < surfels->getNofSurfels(); i++, surfel = surfels->getNextSurfel()) {
				
				newPos = surfel->getPosition() + forces[i];
				gradient = surfel->getNormal();
				if (!this->grid->projectPointOntoDistanceField(newPos, gradient)) {
					newPos = surfel->getPosition();
					gradient = surfel->getNormal();
				}

				// if forces are too strong put the surfel back to origin
				if ((Vector3D(newPos - surfel->getPosition()).getLength()) > (0.5*this->grid->getNarrowBandWidth())) {
					newPos = surfel->getPosition();
					gradient = surfel->getNormal();
				}

				scene->setSurfelPosition (surfel, newPos, false);
				scene->setSurfelNormal(surfel, gradient, false);

				float length = 0.0f;
				float distance = 0.0f;
				int	counter = 0;
				int neighb = 8;
				modelSearch->setNofNeighbours(neighb);
				modelSearch->setSourcePoint(newPos);
				for (j = 1; j < neighb; j++) {
					distance = (Vector3D (surfel->getPosition() - modelSearch->getNeighbourPosition(j))).getLength();
					length += distance;
				}			

				length /= (4*neighb);

				scene->setSurfelRadius(surfel, length, false);
				modelPositions[i] = newPos;
			}	

			this->forceScale *= this->damping;
			scene->emitObjectModifiedSignal (true);
			scene->emitRenderFrameSignal ();
		}
		
		scene->emitObjectModifiedSignal (true);
		scene->emitRenderFrameSignal ();

		delete[] modelPositions;
		delete[] forces;
		delete[] forceScales;
		delete modelSearch;
		this->grid->nodeFactory->freeTraversalNode(travNode);
	}
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
